@@ -21,5 +21,6 @@
  *
  */
 
-char *ouitocomp(const char *oui);
-int oui2comp(const char *oui, char *comp, size_t size);
+#include <bluetooth/bluetooth.h>
+
+char *batocomp(const bdaddr_t *ba);
